# By Justin Walgran
# Copyright (c) 2012 Azavea, Inc.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import os
import re
import platform
import subprocess
from Analyzer import Analyzer
from helpers import first_file_name_in_path_matching_regex


class JsLintComplaint(object):
    def __init__(self, complaint):
        lines = complaint.split('\n')
        regex_match = re.search("line ([0-9]*) character ([0-9]*):(.*)", lines[0])
        if regex_match:
            self.complaint = regex_match.group(2).strip()
            for line in lines:
                self.complaint += "\n" + line
            if len(self.complaint.strip()):
                try:
                    self.line_number = int(regex_match.group(0))
                except:
                    self.line_number = -1
                try:
                    self.column_number = int(regex_match.group(1))
                except:
                    self.column_number = -1
            else:
                self.line_number = -1
                self.column_number = -1
                self.complaint = complaint
        else:
            self.line_number = -1
            self.column_number = -1
            self.complaint = complaint

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        if self.line_number > 0:
            if self.column_number > 0:
                return 'Line %d:%d - %s' % (self.line_number, self.column_number, self.complaint)
            else:
                return 'Line %d - %s' % (self.line_number, self.complaint)
        else:
            return self.complaint


class JSLintAnalyzer(Analyzer):

    def __init__(self, lib_path=None, use_nodejs_if_available=True):
        self._js_lint_proc_args = None
        self._module_path = os.path.dirname(os.path.realpath(__file__))
        self._lib_path = lib_path or os.path.join(self._module_path, 'lib')
        self._jslint_script_regex = re.compile(r'^jslint.*\.js$')
        self._js_lint_script_file_path = first_file_name_in_path_matching_regex(self._lib_path, self._jslint_script_regex)
        self._rhino_jar_regex = re.compile(r'^js\.jar$')
        self._rhino_jar_file_path = first_file_name_in_path_matching_regex(self._lib_path, self._rhino_jar_regex)
        self._lib_message_list = []

        if self._js_lint_proc_args is None and platform.system() == 'Windows':
            # Windows Script Host has two versions, wscript.exe which pops up
            # a window and cscript.exe which does not.
            self._js_lint_proc_args = ["cscript.exe", "//I", "//Nologo", self._js_lint_script_file_path]
            self._lib_message_list.append("Using cscript.exe to run JSLint")

        if self._js_lint_proc_args is None and use_nodejs_if_available:
            null_file = open(os.devnull, 'w')
            try:
                subprocess.check_call(['node', '--help'], stdout=null_file, stderr=null_file)
                self._js_lint_proc_args = ["node", self._js_lint_script_file_path]
                self._lib_message_list.append("Using node.js to run JSLint")
            except Exception:
                self._lib_message_list.append("Cannot use node.js to run JSLint because it was not found on the PATH")
            finally:
                null_file.close()

        if self._js_lint_proc_args is None:
            if self._rhino_jar_file_path:
                self._js_lint_proc_args = ["java", "-jar", self._rhino_jar_file_path, self._js_lint_script_file_path]
                self._lib_message_list.append("Using Rhino to run JSLint")
            else:
                self._lib_message_list.append("Cannot use Rhino to run JSLint because js.jar could not be found in in %r" % self._lib_path)

    def analyze(self, resource):
        analysis = Analyzer.analyze(self, resource)
        analysis.add_messages(self._lib_message_list)
        if self._js_lint_proc_args is None:
            analysis.mark_as_bad()
            analysis.add_error('No suitable JSLint runner (cscript.exe, node.js or rhino) could be found.')
            return analysis

        try:
            js_lint_proc = subprocess.Popen(self._js_lint_proc_args, -1, None, subprocess.PIPE,
                subprocess.PIPE, subprocess.PIPE)
            js_lint_proc_outputs = js_lint_proc.communicate(resource.content)
        except Exception as e:
            if analysis.errors is None:
                analysis.errors = []
            analysis.errors.append("An exception what thrown while running JsLint: " + str(e))
            return analysis

        # The JSLint process returns 1 if it finds lint
        if js_lint_proc.returncode != 0 and js_lint_proc.returncode != 1:
            if analysis.errors is None:
                analysis.errors = []
            analysis.errors.append('The JSLint process exited with return code %d\nArguments: %s\n Output: %s'
                % (js_lint_proc.returncode, self._js_lint_proc_args, js_lint_proc_outputs))
            return analysis

        analysis.mark_as_good()  # Assume that JSLint produced no complaints until parsing one from the process output

        for js_lint_proc_output in js_lint_proc_outputs:
            js_lint_complaints = js_lint_proc_output.split("Lint at ")
            for complaint in js_lint_complaints:
                if len(complaint.strip()):
                    analysis.mark_as_bad()
                    js_lint_complaint = JsLintComplaint(complaint)
                    analysis.add_error(str(js_lint_complaint))

        return analysis



# By Justin Walgran
# Copyright (c) 2012 Azavea, Inc.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import unittest

from blend import Configuration
from blend import Analyzer
from blend.Resource import Resource
from blend.SizeAnalyzer import SizeAnalyzer
from blend import Minifier
from blend.YUICompressorMinifier import YUICompressorMinifier

import os
import shutil
import tempfile

from helpers import clean_output, create_file_with_content


class TestConfiguration(unittest.TestCase):

    def setUp(self):
        self.test_env_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_env_dir)
        clean_output()

    def test_can_add_analyzer_for_filetype(self):
        conf = Configuration()
        analyzer = Analyzer()
        conf.add_analyzer_for_file_type(analyzer, 'javascript')
        resource = Resource('file.js')
        analyzers = conf.get_analyzers_for_resource(resource)
        self.assertListEqual([analyzer], analyzers)

    def test_returns_non_when_asking_for_analyzers_for_an_unknown_file_type(self):
        conf = Configuration()
        analyzer = Analyzer()
        conf.add_analyzer_for_file_type(analyzer, 'javascript')
        resource = Resource('file.foo')
        analyzers = conf.get_analyzers_for_resource(resource)
        self.assertIsNone(analyzers)

    def test_add_analyzer_checks_classes(self):
        conf = Configuration()
        self.assertRaises(Exception, conf.add_analyzer_for_file_type, 'string instead of an analyzer', 'javascript')

        # should not throw
        conf.add_analyzer_for_file_type(Analyzer(), 'javascript')
        # should not throw
        conf.add_analyzer_for_file_type(SizeAnalyzer(), 'javascript')

    def test_throws_when_passed_an_invalid_config_file_path(self):
        self.assertRaises(Exception, Configuration, '/some/non/existent/path')

    def test_can_load_analyzers_from_config_file(self):
        config_file_path = os.path.join(self.test_env_dir, 'blend.config')
        create_file_with_content(config_file_path,
"""{
    "analyzers": {
        "javascript": [
            {
                "name": "blend.SizeAnalyzer",
                "skip_list": [
                    "bin"
                ]
            }
        ]
    }
}""")
        conf = Configuration(config_file_path)
        resource = Resource('file.js')
        actual_analyzers = conf.get_analyzers_for_resource(resource)
        self.assertIsNotNone(actual_analyzers)
        self.assertEqual(1, len(actual_analyzers))
        self.assertIsInstance(actual_analyzers[0], SizeAnalyzer)
        self.assertIsNotNone(conf.analyzer_skip_lists)

    def test_can_load_minfiers_from_config_file(self):
        config_file_path = os.path.join(self.test_env_dir, 'blend.config')
        create_file_with_content(config_file_path,
            """{
                "minifiers": {
                    "javascript": {
                        "name": "blend.YUICompressorMinifier"
                    }
                }
            }""")
        conf = Configuration(config_file_path)
        resource = Resource('file.js')
        actual_minifier = conf.get_minifier_for_file_type(resource.file_type)
        self.assertIsNotNone(actual_minifier)
        self.assertIsInstance(actual_minifier, YUICompressorMinifier)

    def test_can_add_minifier_for_filetype(self):
        conf = Configuration()
        minifier = Minifier()
        conf.set_minifier_for_file_type(minifier, 'javascript')
        actual_minifier = conf.get_minifier_for_file_type('javascript')
        self.assertEqual(minifier, actual_minifier)

    def test_add_minifier_checks_classes(self):
        conf = Configuration()
        self.assertRaises(Exception, conf.set_minifier_for_file_type, 'string instead of an minifier', 'javascript')

        # should not throw
        conf.set_minifier_for_file_type(Minifier(), 'javascript')
        # should not throw
        conf.set_minifier_for_file_type(YUICompressorMinifier(), 'javascript')

    def test_returns_none_when_asking_for_minifier_for_an_unknown_file_type(self):
        conf = Configuration()
        minifier = Minifier()
        conf.set_minifier_for_file_type(minifier, 'javascript')
        analyzers = conf.get_minifier_for_file_type('some-other-type')
        self.assertIsNone(analyzers)

    def test_get_analyzers_for_resource_with_skip_list(self):
        lib_resource = Resource(os.path.join(os.getcwd(), 'lib', 'jquery.js'))
        src_resource = Resource(os.path.join(os.getcwd(), 'src', 'file.js'))
        conf = Configuration()
        analyzer = Analyzer()
        conf.add_analyzer_for_file_type(analyzer, 'javascript', ['lib/*'])
        self.assertIsNone(conf.get_analyzers_for_resource(lib_resource))
        self.assertEqual([analyzer], conf.get_analyzers_for_resource(src_resource))

    def test_add_analyzer_for_file_type_raises_when_skip_list_is_a_string(self):
        conf = Configuration()
        self.assertRaises(Exception, conf.add_analyzer_for_file_type, Analyzer(), 'javascript', 'something invalid')



# By Justin Walgran
# Copyright (c) 2012 Azavea, Inc.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import optparse
import os
import sys
import traceback
from Paths import Paths
from Resource import Resource
from Configuration import Configuration
from JSLintAnalyzer import JSLintAnalyzer
from YUICompressorMinifier import YUICompressorMinifier
from blend.Requirement import RequirementNotSatisfiedException


class Application():
    DEFAULT_OUTPUT_DIR = os.path.join(os.getcwd(), 'output')
    DEFAULT_PATH_LIST = []
    DEFAULT_INCLUDE_CWD = True
    DEFAULT_FILE_LIST = []
    DEFAULT_CONFIG_FILE_PATH = os.path.join(os.getcwd(), '.blend', 'config.json')

    def __init__(self, path_list=DEFAULT_PATH_LIST, include_cwd=DEFAULT_INCLUDE_CWD,
                 file_list=DEFAULT_FILE_LIST, output_dir=DEFAULT_OUTPUT_DIR, config_file_path=DEFAULT_CONFIG_FILE_PATH):
        self.paths = Paths(*path_list, include_cwd=include_cwd)
        self.include_cwd = include_cwd
        self.file_list = file_list
        self.output_dir = output_dir
        if os.path.exists(config_file_path):
            print config_file_path
            self.config = Configuration(config_file_path)
        else:
            self.config = self._create_default_configuration()

    def _create_default_configuration(self):
        config = Configuration()
        config.add_analyzer_for_file_type(JSLintAnalyzer(), 'javascript', [os.path.join('lib', '*')])
        config.set_minifier_for_file_type(YUICompressorMinifier(), 'javascript')
        config.set_minifier_for_file_type(YUICompressorMinifier(), 'css')
        return config

    def run(self):
        try:
            if not os.path.exists(self.output_dir):
                os.makedirs(self.output_dir)

            if len(self.file_list) == 0:
                resources = Resource.find_all_in_paths(self.paths)
            else:
                resources = []
                for file_path in self.file_list:
                    resources.append(Resource(file_path))
            if resources is not None:
                for resource in resources:
                    if resource.requirements is not None:
                        directory, file_name = os.path.split(resource.path_to_file)

                        try:
                            chunks = resource.get_chunks_by_merging_requirements_from_paths(self.paths, previously_merged=[])
                        except RequirementNotSatisfiedException, rnse:
                            print "A requirement could not be satisfied for %s\n\n%s\n" % (resource.path_to_file, rnse)
                            return -1

                        for chunk in chunks:
                            analyzers = self.config.get_analyzers_for_resource(chunk.resource)
                            if analyzers:
                                for analyzer in analyzers:
                                    print 'Analysis:%s:%s' % (analyzer.__class__, chunk.resource.path_to_file)
                                    analysis = analyzer.analyze(chunk.resource)
                                    print analysis
                                    if not analysis.good:
                                        return -1

                        merged_content = ''.join([chunk.content for chunk in chunks])

                        output_file_name = os.path.join(self.output_dir, file_name)
                        f = open(output_file_name, 'w')
                        try:
                            f.write(merged_content)
                        finally:
                            f.flush()
                            f.close()

                        print "Created %s" % output_file_name

                        # TODO: Process chunks to prevent reminification
                        output_resource = Resource(output_file_name)
                        minifier = self.config.get_minifier_for_file_type(output_resource.file_type)
                        if minifier and not output_resource.minified:
                            minification = minifier.minify(output_resource)
                            if not minification.good:
                                print minification
                                return -1
                            minified_output_file_path = os.path.join(self.output_dir, output_resource.minified_file_name)
                            f = open(minified_output_file_path, 'w')
                            try:
                                f.write(minification.content)
                            finally:
                                f.flush()
                                f.close()

        except Exception:
            traceback.print_exc(file=sys.stderr)
            return -1

        return 0

    @staticmethod
    def main():
        parser = optparse.OptionParser("""usage %prog [options] [file1 [file2 [fileN]]]

If no file arguments are specified, blend searches the specified
paths and processes any and all files that require other files.

If no paths are specified, the current working directory is
searched for required files.

If no output path is specified, an output directory is created in the
current working directory and all outputs are written to that directory.""")

        parser.add_option("-o", "--output",
            default=Application.DEFAULT_OUTPUT_DIR,
            dest='output_dir',
            metavar='OUTPUT',
            help="where the file output will be written")

        parser.add_option("-p", "--path",
            default=Application.DEFAULT_PATH_LIST,
            dest='path',
            metavar='PATH',
            action='append',
            help='a directory to be searched for required files (multiple directories can specified by repeating the flag)')

        parser.add_option("-s", "--skipcwd",
            default=not Application.DEFAULT_INCLUDE_CWD,
            dest='skip_cwd',
            metavar='ENV',
            action='store_true',
            help='exclude the current working directory from the search path')

        parser.add_option("-c", "--config",
            default=Application.DEFAULT_CONFIG_FILE_PATH,
            dest='config_file_path',
            metavar='CONFIG',
            help='the JSON format config file from which to load application settings')

        options, arguments = parser.parse_args()

        file_list = arguments or []

        app = Application(options.path, not options.skip_cwd, file_list, options.output_dir, options.config_file_path)
        sys.exit(app.run())

if __name__ == '__main__':
    Application.main()



# By Justin Walgran
# Copyright (c) 2012 Azavea, Inc.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

from Analyzer import Analyzer


class SizeAnalyzer(Analyzer):

    def analyze(self, resource):
        analysis = Analyzer.analyze(self, resource)
        # This analyzer only computes the size of the content, it does not judge the quality
        lines = resource.content.split('\n')
        analysis.mark_as_good()
        if len(lines) == 1:
            line_noun = 'line'
        else:
            line_noun = 'lines'
        char_count = reduce(lambda count, x: count + len(x), lines, 0)
        analysis.add_message('%s: %d characters in %d %s for %d bytes' % (resource.path_to_file, char_count, len(lines), line_noun, resource.size))

        return analysis



# By Justin Walgran
# Copyright (c) 2012 Azavea, Inc.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import os
import shutil


def create_files(paths_to_files):
    if isinstance(paths_to_files, basestring):
        internal_paths_to_files = [paths_to_files]
    else:
        internal_paths_to_files = paths_to_files
    for path_to_file in internal_paths_to_files:
        if not os.path.exists(path_to_file):
            if os.path.dirname(path_to_file) != '' and not os.path.exists(os.path.dirname(path_to_file)):
                os.makedirs(os.path.dirname(path_to_file))
            open(path_to_file, 'w').close()


def create_file_with_content(path_to_test_file, content):
    create_files(path_to_test_file)
    f = open(path_to_test_file, 'w')
    try:
        f.write(content)
    finally:
        f.close()


def clean_up_files(paths_to_files):
    if isinstance(paths_to_files, basestring):
        internal_paths_to_files = [paths_to_files]
    else:
        internal_paths_to_files = paths_to_files
    for path_to_file in internal_paths_to_files:
        if os.path.exists(path_to_file):
            os.remove(path_to_file)


def clean_output():
    output_dir = os.path.join(os.getcwd(), "output")
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)



# By Justin Walgran
# Copyright (c) 2012 Azavea, Inc.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.


class Requirement:
    """
    Represents the a dependency that one Resource has on another.
    """
    def __init__(self, name, insert_location=0):
        """
        Arguments:
        name -- The string name of the required resource
        insert_location -- The character position within the content where the required
        resource will be inserted during a merge. Can be either a single integer for
        inserting without replacement or a tuple defining the character range in the
        requiring resource that will be replaced by the content of the required resource.
        """
        if not name:
            raise Exception('Requirement must be created with a name')
        if not isinstance(name, str):
            raise Exception('The name argument must be set to a string')
        self._name = name

        if isinstance(insert_location, basestring):
            raise Exception('The insert_location argument must be a number or a two-number tuple')
        self._insert_location = insert_location

    @property
    def name(self):
        """
        The string name of the required resource
        """
        return self._name

    @property
    def standard_name(self):
        """
        The lowercase name of the required resource
        """
        return self._name.lower()

    @property
    def insert_location(self):
        """
        The character position within the content where the required resource will be
        inserted during a merge. Can be either a single integer for inserting without
        replacement or a tuple defining the character range in the requiring resource
        that will be replaced by the content of the required resource.
        """
        return self._insert_location


class RequirementNotSatisfiedException(Exception):
    def __init__(self, requirement, paths, *args, **kwargs):
        super(RequirementNotSatisfiedException, self).__init__(*args, **kwargs)
        self.requirement = requirement
        self.paths = paths

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return "Requirement:\n\t%s\nSearch paths:\n\t%s" % (self.requirement.name, '\n\t'.join(self.paths.search_paths))



# By Justin Walgran
# Copyright (c) 2012 Azavea, Inc.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import unittest
import tempfile

import shutil
import os
import sys
from blend import Application
from helpers import create_files, create_file_with_content, clean_up_files, clean_output


class TestApplication(unittest.TestCase):

    def setUp(self):
        self.test_env_dir = tempfile.mkdtemp()
        self.test_config_file_path = os.path.join(self.test_env_dir, 'config.json')
        self.default_config_file_path = os.path.join(os.getcwd(), '.blend', 'config.json')
        if os.path.exists(self.default_config_file_path):
            shutil.rmtree(os.path.dirname(self.default_config_file_path))

    def tearDown(self):
        shutil.rmtree(self.test_env_dir)
        if os.path.exists(self.default_config_file_path):
            shutil.rmtree(os.path.dirname(self.default_config_file_path))
        clean_output()

    def test_default_env_is_cwd(self):
        app = Application()
        self.assertEquals(app.paths.search_paths, [os.getcwd()])

    def test_default_include_cwd_is_true(self):
        app = Application()
        self.assertTrue(app.include_cwd)

    def test_default_file_list_is_empty(self):
        app = Application()
        self.assertEqual([], app.file_list)

    def test_default_output_dir_is_cwd_plus_output(self):
        app = Application()
        self.assertEqual(os.path.join(os.getcwd(), 'output'),  app.output_dir)

    def test_run_without_arguments_returns_zero(self):
        app = Application()
        self.assertEqual(0, app.run())

    def test_run_with_single_resource(self):
        paths_to_processable_test_files = [
            os.path.join(self.test_env_dir, 'dir1', 'file1.js')]
        create_files(paths_to_processable_test_files)
        app = Application(file_list=[paths_to_processable_test_files[0]])
        self.assertEqual(0, app.run())
        clean_up_files(paths_to_processable_test_files)

    def test_run_with_single_resource_with_requirement(self):
        paths_to_test_files = [
            os.path.join(self.test_env_dir, 'dir1', 'file1.js'),
            os.path.join(self.test_env_dir, 'dir2', 'file2.js')]
        clean_up_files(paths_to_test_files)
        create_files(paths_to_test_files)
        create_file_with_content(paths_to_test_files[0], '// This is file 1\n//= require file2')
        create_file_with_content(paths_to_test_files[1], '// This is file 2')
        app = Application(path_list=[self.test_env_dir], file_list=[paths_to_test_files[0]])
        self.assertEqual(0, app.run())
        clean_up_files(paths_to_test_files)

    @unittest.skipIf(len(sys.argv) > 1, "If arguments are passed to the unit test runner, this test fails")
    def test_main_exits_cleanly_when_no_args_are_passed(self):
        app = Application
        try:
            app.main()
        except SystemExit, system_exit_exception:
            self.assertEquals(0, system_exit_exception.code)

    def test_application_can_be_created_with_a_config_file_path(self):
        clean_up_files(self.test_config_file_path)
        create_file_with_content(self.test_config_file_path, '{}')
        Application(config_file_path=self.test_config_file_path)

    def test_loads_config_from_file(self):
        # To ensure that somthing is actually being tested, I first assert that the default configuration
        # is non-empty
        app = Application()
        self.assertTrue(app.config.analyzers, "expected the default analyzer list to be non-empty")
        self.assertTrue(app.config.minifiers, "expected the default minifier list to be non-empty")

        clean_up_files(self.test_config_file_path)
        create_file_with_content(self.test_config_file_path, '{}')
        app = Application(config_file_path=self.test_config_file_path)
        self.assertIsNotNone(app.config)
        self.assertFalse(app.config.analyzers, "expected the analyzer list to be empty")
        self.assertFalse(app.config.minifiers, "expected the minifier list to be empty")

    def test_loads_config_from_default_file_in_the_dot_blend_directory(self):
        # To ensure that somthing is actually being tested, I first assert that the default configuration
        # is non-empty
        app = Application()
        self.assertTrue(app.config.analyzers, "expected the default analyzer list to be non-empty")
        self.assertTrue(app.config.minifiers, "expected the default minifier list to be non-empty")

        clean_up_files(self.default_config_file_path)
        create_file_with_content(self.default_config_file_path, '{}')

        app = Application()
        self.assertIsNotNone(app.config)
        self.assertFalse(app.config.analyzers, "expected the analyzer list to be empty")
        self.assertFalse(app.config.minifiers, "expected the minifier list to be empty")

    def tests_config_file_passed_as_an_argument_overrides_the_file_in_dot_blend(self):
        # Step 1: Assert the default configuration with no config files
        app = Application()
        self.assertTrue(app.config.analyzers, "expected the default analyzer list to be non-empty")
        self.assertTrue(app.config.minifiers, "expected the default minifier list to be non-empty")

        # Step 2: Assert that config is loaded from .blend/config.json
        clean_up_files(self.default_config_file_path)
        create_file_with_content(self.default_config_file_path, '{}')

        app = Application()
        self.assertIsNotNone(app.config)
        self.assertFalse(app.config.analyzers, "expected the analyzer list to be empty")
        self.assertFalse(app.config.minifiers, "expected the minifier list to be empty")

        # Step 3: Assert that a custom config file overrides .blend/config.json
        clean_up_files(self.test_config_file_path)
        create_file_with_content(self.test_config_file_path, """{
    "analyzers": {
        "javascript": [
            {
                "name": "blend.SizeAnalyzer",
                "skip_list": [
                    "bin"
                ]
            }
        ]
    }
}""")
        app = Application(config_file_path=self.test_config_file_path)
        self.assertIsNotNone(app.config)
        self.assertEqual(1, len(app.config.analyzers), "expected one analyzer")
        self.assertFalse(app.config.minifiers, "expected the minifier list to be empty")



# By Justin Walgran
# Copyright (c) 2012 Azavea, Inc.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

from Resource import Resource
from Requirement import Requirement
from Paths import Paths
from Application import Application
from Analyzer import Analyzer
from Analysis import Analysis
from SizeAnalyzer import SizeAnalyzer
from Configuration import Configuration
from JSLintAnalyzer import JSLintAnalyzer
from Minifier import Minifier
from Result import Result
from Minification import Minification
from YUICompressorMinifier import YUICompressorMinifier

__version__ = '0.0.1'


# By Justin Walgran
# Copyright (c) 2012 Azavea, Inc.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import os
import re
from copy import deepcopy
import sys

from Requirement import Requirement, RequirementNotSatisfiedException


class Resource:
    """
    Representation of a file on disk
    """
    def __init__(self, path_to_file):
        """
        Arguments:
        path_to_file -- The path at which the physical file is/will be located.
        """
        if not path_to_file:
            raise Exception('Resource must be created with a path_to_file')
        if not isinstance(path_to_file, str):
            raise Exception('The path_to_file argument must be set to a string')

        self._path_to_file = path_to_file
        self._extension, self._file_type = Resource._parse_extension_and_file_type(path_to_file)
        self._base_name, self._minified = Resource._parse_base_name_and_minification_status(path_to_file)
        self._set_content_and_size(path_to_file)
        self._set_requirements()

    def _set_content_and_size(self, path_to_file):
        """
        Reads the content of the file specified by path_to_file into the _content
        member variable.
        Arguments:
        path_to_file -- The path at which the physical file is be located.
        Remarks:
        If path_to_file specifies a non-existent file the _content member variable is
        set to None
        """
        if os.path.exists(path_to_file):
            self._size = os.path.getsize(path_to_file)
            f = open(path_to_file, 'r')
            try:
                self._content = f.read()
            finally:
                f.close()
        else:
            self._content = None
            self._size = 0

    def _set_requirements(self):
        """
        Parse the content of the file and set the _requirements member variable to a
        list Requirement objects.
        Remarks:
        Must be called AFTER _set_content() is called and after self._file_type is set.
        """
        self._requirements = None

        if self._file_type == 'unknown' or self._content is None:
            return

        if self._file_type == 'javascript':
            require_re = re.compile(r'[ \t]*//=[ \t]*require[ \t]+(\S+)[ \t]*\n?')
        else:  # 'css'
            require_re = re.compile(r'@import url\(\"(?P<import_match>\S+)\.\S+"\)|[ \t]*/\*=[ \t]*require[ \t]+(?P<require_match>\S+)[ \t]*\*/\n?')

        # This loop expects each match to have 2 groups. The first group captures the
        # punctuation mark around the file name and the second captures the name itself
        for result in require_re.finditer(self._content):
            if self._file_type == 'javascript':
                name = str(result.groups()[0])
            else:
                if result.group('import_match') is not None:
                    name = str(result.group('import_match'))
                else:
                    name = str(result.group('require_match'))

            insert_position = (result.start(), result.end())
            if self._requirements is None:
                self._requirements = []
            self._requirements.append(Requirement(name, insert_position))

    @staticmethod
    def _parse_extension_and_file_type(path_to_file):
        """
        Extract a lower case extension from the specified file name and return it
        along with a file type string
        Arguments:
        path_to_file -- The full path to a file that may or may not exist.
        """
        base_name, ext = os.path.splitext(path_to_file)
        ext = ext.lower()[1:]

        if ext == 'js' or ext == 'javascript':
            file_type = 'javascript'
        elif ext == 'css':
            file_type = 'css'
        else:
            file_type = 'unknown'

        return ext, file_type

    @staticmethod
    def _parse_base_name_and_minification_status(path_to_file):
        """
        Extract a lower case name from the specified path_to_file by removing the
        extension, '-min', and and version number if they are present.
        Arguments:
        path_to_file -- The full path to a file that may or may not exist.
        """
        directory, file = os.path.split(path_to_file)
        name, dot, extension = file.rpartition('.')
        lower_name = name.lower()
        minified = False
        if lower_name[-4:] == '-min' or lower_name[-4:] == '.min':
            lower_name = lower_name[:-4]
            minified = True
        lower_name_without_version, dash, version = lower_name.rpartition('-')
        if dash == '':
            return lower_name, minified
        else:
            return lower_name_without_version, minified

    @property
    def path_to_file(self):
        """
        The path at which the physical file is/will be located.
        """
        return self._path_to_file

    @property
    def exists(self):
        """
        Whether or not the file represented by the Resource exists on disk.
        """
        return os.path.exists(self.path_to_file)

    @property
    def size(self):
        """
        The path at which the physical file is/will be located.
        """
        return self._size

    @property
    def content(self):
        """
        The contents of the file.
        """
        return self._content

    @property
    def extension(self):
        """
        The file name extension of the physical file in lower case without the '.' character.
        """
        return self._extension

    @property
    def file_type(self):
        """
        A lower case string describing the content of this file.
        Possible values: unknown, javascript, css.
        """
        return self._file_type

    @property
    def base_name(self):
        """
        The name of the resource with version numbers and minification designations removed.
        Example: OpenLayers.js -> openlayers
        Example: jQuery-1.5.4-min.js -> jquery
        """
        return self._base_name

    @property
    def minified(self):
        return self._minified

    @property
    def minified_file_name(self):
        return '%s-min.%s' % (self.base_name, self.extension)

    @property
    def requirements(self):
        """
        The descriptions of the the other resources on which this resource depends.
        """
        return self._requirements

    def get_chunks_by_merging_requirements_from_paths(self, paths, previously_merged):
        if self.requirements:
            map = self.map_requirements(paths, map={}, previously_required=[])

            chunks_and_requirements = []
            position = 0

            for requirement in self.requirements:
                if (position == 0 and requirement.insert_location[0] > 0) or position != requirement.insert_location[0]:
                    chunks_and_requirements.append(Chunk(self, position, requirement.insert_location[0]))
                chunks_and_requirements.append(requirement)
                position = requirement.insert_location[1]
            if position < len(self.content):
                chunks_and_requirements.append(Chunk(self, position))

            chunks = []
            for chunk_or_requirement in chunks_and_requirements:
                if isinstance(chunk_or_requirement, Requirement):
                    resource = map[chunk_or_requirement.standard_name]['resource']
                    if resource.base_name not in previously_merged:
                        chunks = chunks + resource.get_chunks_by_merging_requirements_from_paths(paths, previously_merged)
                    previously_merged.append(resource.base_name)
                else:
                    chunks.append(chunk_or_requirement)

            return chunks
        else:
            return [Chunk(self)]

    def merge_requirements_from_paths(self, paths, previously_merged):
        chunks = self.get_chunks_by_merging_requirements_from_paths(paths, previously_merged)
        return ''.join([chunk.content for chunk in chunks])

    def distance_to_file(self, path2):
        abspath1 = os.path.abspath(self.path_to_file)
        abspath2 = os.path.abspath(path2)

        # If the files are in the same directory, return the smallest possible
        # integer as the distance.
        if os.path.dirname(abspath1) == os.path.dirname(abspath2):
            return -1 * sys.maxint

        downward_distance = 0
        dir2_head = os.path.dirname(abspath2)
        dir2_tail = None
        while dir2_tail != '':
            dir1_head = os.path.dirname(os.path.abspath(self.path_to_file))
            dir1_tail = None
            upward_distance = 0
            while dir1_tail != '':
                if dir2_head == dir1_head:
                    # Lower distance is always better, and downward is always preferable to upward.
                    # The weight of 100000 applied to the upward distance is arbitrary, but practically
                    # ensures that a downward match always has a lower distance in all practical cases.
                    return (-1 * sys.maxint) + downward_distance + (upward_distance * 100000)
                dir1_head, dir1_tail = os.path.split(dir1_head)
                upward_distance += 1
            dir2_head, dir2_tail = os.path.split(dir2_head)
            downward_distance += 1

        # If the two paths do not share anything in common, return
        # the largest possible distance.
        return sys.maxint

    def map_requirements(self, paths, map, previously_required):
        if self.requirements:
            for requirement in self.requirements:
                map[requirement.standard_name] = None

                resources_of_the_same_type = Resource.find_all_of_type_in_paths(self.file_type, paths)

                if resources_of_the_same_type is not None:
                    matching_resources = []
                    for resource in resources_of_the_same_type:
                        if resource.base_name == requirement.standard_name:
                            matching_resources.append(resource)

                    matching_resources_in_order = sorted(matching_resources,
                        key=lambda item: self.distance_to_file(item.path_to_file))

                    if len(matching_resources_in_order) > 0:
                        resource = matching_resources_in_order[0]
                        new_previously_required = deepcopy(previously_required)
                        new_previously_required.append(requirement.standard_name)
                        resource.map_requirements(paths, map, new_previously_required)

                        map[requirement.standard_name] = {
                            'resource': resource,
                            'previously_required': deepcopy(previously_required)
                        }

                if not map[requirement.standard_name]:
                    raise RequirementNotSatisfiedException(requirement, paths)

        return map

    @staticmethod
    def find_all_of_type_in_paths(file_type, paths):
        """
        Get a list of Resource instances representing all the files in the current working
        directory that have the specified file_type
        Arguments:
        file_type -- The string name of the type of file to be found. Can be unknown, javascript, or css
        paths -- A Paths instance defining where to search for files.
        Remarks:
        Calls find_all_of_type_in_path for each path in the paths.
        """
        resources = []
        for path in paths.search_paths:
            resources.extend(Resource.find_all_of_type_in_path(file_type, path, skip_path=paths.output_path))
        return resources if len(resources) > 0 else None

    @staticmethod
    def find_all_of_type_in_path(file_type, path, skip_path=None):
        """
        Get a list of Resource instances representing all the files in the specified
        directory (and sub-directories) that have the specified file_type
        Arguments:
        file_type -- The string name of the type of file to be found. Can be unknown, javascript, or css.
        path -- The base directory to be recursively searched for files.
        """
        resources = []
        for dir_path, dir_names, file_names in os.walk(path):
            if skip_path is None or os.path.abspath(dir_path) != os.path.abspath(skip_path):
                for file_name in file_names:
                    resource = Resource(os.path.join(dir_path, file_name))
                    if resource.file_type == file_type:
                        resources.append(resource)
        return resources

    @staticmethod
    def find_all_in_paths(paths):
        """
        Get a list of Resource instances representing all the files in the specified
        paths that have a processable file type.
        Arguments:
        paths -- A Paths instance defining where to search for files.
        """
        resources = []
        for path in paths.search_paths:
            resources_in_path = []
            for dir_path, dir_names, file_names in os.walk(path):
                for file_name in file_names:
                    absolute_file_path = os.path.join(dir_path, file_name)
                    ext, file_type = Resource._parse_extension_and_file_type(absolute_file_path)
                    if file_type != 'unknown':
                        resources_in_path.append(Resource(absolute_file_path))
            resources.extend(resources_in_path)
        return resources if len(resources) > 0 else None

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return self.base_name


class Chunk():
    def __init__(self, resource, start=None, end=None):
        self._resource = resource
        self._start = start
        self._end = end

    @property
    def resource(self):
        return self._resource

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        value = self.resource.base_name
        if self.start is not None:
            value = value + ':' + self.start
        if self.end is not None:
            value = value + ':' + self.end
        return value

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end

    @property
    def content(self):
        return self._resource.content[self._start:self._end]
